from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Union

import plotext as plt
from dolphie.Modules.Functions import format_bytes, format_number, format_time
from rich.text import Text
from textual.widgets import Static


class Graph(Static):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.marker = None
        self.metric_instance = None

    def on_show(self) -> None:
        self.render_graph(self.metric_instance)

    def on_resize(self) -> None:
        self.render_graph(self.metric_instance)

    def render_graph(self, metric_instance) -> None:
        self.metric_instance = metric_instance

        if self.metric_instance is None:
            return

        plt.clf()
        plt.date_form("d/m/y H:M:S")
        plt.canvas_color((10, 14, 27))
        plt.axes_color((10, 14, 27))
        plt.ticks_color((133, 159, 213))

        plt.plotsize(self.size.width, self.size.height)

        max_y_value = 0
        if isinstance(self.metric_instance, CheckpointMetrics):
            x = self.metric_instance.datetimes
            y = self.metric_instance.Innodb_checkpoint_age.values

            if y:
                plt.hline(0, (10, 14, 27))
                plt.hline(self.metric_instance.checkpoint_age_sync_flush, (241, 251, 130))
                plt.hline(self.metric_instance.checkpoint_age_max, (252, 121, 121))
                plt.text(
                    "Critical",
                    y=self.metric_instance.checkpoint_age_max,
                    x=max(x),
                    alignment="right",
                    color=(233, 233, 233),
                    style="bold",
                )
                plt.text(
                    "Warning",
                    y=self.metric_instance.checkpoint_age_sync_flush,
                    x=max(x),
                    alignment="right",
                    color=(233, 233, 233),
                    style="bold",
                )

                plt.plot(
                    x,
                    y,
                    marker=self.marker,
                    label=self.metric_instance.Innodb_checkpoint_age.label,
                    color=self.metric_instance.Innodb_checkpoint_age.color,
                )
                max_y_value = self.metric_instance.checkpoint_age_max
        elif isinstance(self.metric_instance, RedoLogMetrics) and "graph_redo_log_bar" in self.id:
            if self.metric_instance.Innodb_lsn_current.values:
                x = [0]
                y = [
                    round(
                        sum(self.metric_instance.Innodb_lsn_current.values)
                        * (3600 / len(self.metric_instance.Innodb_lsn_current.values))
                    )
                ]

                plt.hline(self.metric_instance.redo_log_size, (252, 121, 121))
                plt.text(
                    "Log Size",
                    y=self.metric_instance.redo_log_size,
                    x=0,
                    alignment="center",
                    color=(233, 233, 233),
                    style="bold",
                )

                bar_color = (46, 124, 175)
                if y[0] >= self.metric_instance.redo_log_size:
                    bar_color = (252, 121, 121)

                plt.text(
                    format_bytes(y[0], color=False) + "/hr",
                    y=y[0],
                    x=0,
                    alignment="center",
                    color=(233, 233, 233),
                    style="bold",
                    background=bar_color,
                )

                plt.bar(
                    x,
                    y,
                    marker="hd",
                    color=bar_color,
                )
                max_y_value = max(self.metric_instance.redo_log_size, max(y))
        elif isinstance(self.metric_instance, RedoLogActiveCountMetrics):
            x = self.metric_instance.datetimes
            y = self.metric_instance.Active_redo_log_count.values

            if y:
                plt.hline(1, (10, 14, 27))
                plt.hline(34, (252, 121, 121))
                plt.text(
                    "Max Count",
                    y=34,
                    x=max(x),
                    alignment="right",
                    color=(233, 233, 233),
                    style="bold",
                )

                plt.plot(
                    x,
                    y,
                    marker=self.marker,
                    label=self.metric_instance.Active_redo_log_count.label,
                    color=self.metric_instance.Active_redo_log_count.color,
                )
                max_y_value = 32

        else:
            for metric_data in self.metric_instance.__dict__.values():
                if isinstance(metric_data, MetricData) and metric_data.visible:
                    x = self.metric_instance.datetimes
                    y = metric_data.values

                    if y:
                        plt.plot(x, y, marker=self.marker, label=metric_data.label, color=metric_data.color)
                        max_y_value = max(max_y_value, max(y))

        max_y_ticks = 5
        y_tick_interval = max_y_value / max_y_ticks

        if y_tick_interval >= 1:
            y_ticks = [i * y_tick_interval for i in range(int(max_y_ticks) + 1)]
        else:
            y_ticks = [i for i in range(int(max_y_value) + 1)]

        format_function = get_number_format_function(self.metric_instance)
        y_labels = [format_function(val) for val in y_ticks]

        plt.yticks(y_ticks, y_labels)

        # Fix for Windows when graphs initially load
        try:
            self.update(Text.from_ansi(plt.build()))
        except OSError:
            pass


def get_number_format_function(data, color=False):
    data_formatters = {
        ReplicationLagMetrics: lambda val: format_time(val),
        CheckpointMetrics: lambda val: format_bytes(val, color=color),
        RedoLogMetrics: lambda val: format_bytes(val, color=color),
        AdaptiveHashIndexHitRatioMetrics: lambda val: f"{round(val)}%",
        DiskIOMetrics: lambda val: format_bytes(val, color=color),
    }

    return data_formatters.get(type(data), lambda val: format_number(val, color=color))


@dataclass
class MetricSource:
    global_status: str = "global_status"
    innodb_metrics: str = "innodb_metrics"
    disk_io_metrics: str = "disk_io_metrics"
    none: str = "none"


@dataclass
class MetricColor:
    gray: tuple = (172, 207, 231)
    blue: tuple = (68, 180, 255)
    green: tuple = (84, 239, 174)
    red: tuple = (255, 73, 112)
    yellow: tuple = (252, 213, 121)


@dataclass
class MetricData:
    label: str
    color: tuple
    visible: bool = True
    save_history: bool = True
    per_second_calculation: bool = True
    last_value: int = None
    graphable: bool = True
    values: List[int] = field(default_factory=list)


@dataclass
class DMLMetrics:
    Queries: MetricData
    Com_select: MetricData
    Com_insert: MetricData
    Com_update: MetricData
    Com_delete: MetricData
    Com_replace: MetricData
    Com_commit: MetricData
    Com_rollback: MetricData
    graphs: List[str]
    tab_name: str = "dml"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class ReplicationLagMetrics:
    lag: MetricData
    graphs: List[str]
    tab_name: str = "replication_lag"
    metric_source: MetricSource = MetricSource.none

    datetimes: List[str] = field(default_factory=list)


@dataclass
class CheckpointMetrics:
    Innodb_checkpoint_age: MetricData
    graphs: List[str]
    tab_name: str = "checkpoint"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)
    checkpoint_age_max: int = 0
    checkpoint_age_sync_flush: int = 0


@dataclass
class BufferPoolRequestsMetrics:
    Innodb_buffer_pool_read_requests: MetricData
    Innodb_buffer_pool_write_requests: MetricData
    Innodb_buffer_pool_reads: MetricData
    graphs: List[str]
    tab_name: str = "buffer_pool_requests"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class AdaptiveHashIndexMetrics:
    adaptive_hash_searches: MetricData
    adaptive_hash_searches_btree: MetricData
    graphs: List[str]
    tab_name: str = "adaptive_hash_index"
    metric_source: MetricSource = MetricSource.innodb_metrics
    datetimes: List[str] = field(default_factory=list)


@dataclass
class AdaptiveHashIndexHitRatioMetrics:
    hit_ratio: MetricData
    graphs: List[str]
    smoothed_hit_ratio: float = None
    tab_name: str = "adaptive_hash_index"
    metric_source: MetricSource = MetricSource.none
    datetimes: List[str] = field(default_factory=list)


@dataclass
class RedoLogMetrics:
    Innodb_lsn_current: MetricData
    graphs: List[str]
    tab_name: str = "redo_log"
    redo_log_size: int = 0
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class RedoLogActiveCountMetrics:
    Active_redo_log_count: MetricData
    graphs: List[str]
    tab_name: str = "redo_log"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class TableCacheMetrics:
    Table_open_cache_hits: MetricData
    Table_open_cache_misses: MetricData
    Table_open_cache_overflows: MetricData
    graphs: List[str]
    tab_name: str = "table_cache"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class ThreadMetrics:
    Threads_connected: MetricData
    Threads_running: MetricData
    graphs: List[str]
    tab_name: str = "threads"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class TemporaryObjectMetrics:
    Created_tmp_tables: MetricData
    Created_tmp_disk_tables: MetricData
    Created_tmp_files: MetricData
    graphs: List[str]
    tab_name: str = "temporary_objects"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class AbortedConnectionsMetrics:
    Aborted_clients: MetricData
    Aborted_connects: MetricData
    graphs: List[str]
    tab_name: str = "aborted_connections"
    metric_source: MetricSource = MetricSource.global_status
    datetimes: List[str] = field(default_factory=list)


@dataclass
class DiskIOMetrics:
    io_read: MetricData
    io_write: MetricData
    graphs: List[str]
    tab_name: str = "disk_io"
    metric_source: MetricSource = MetricSource.disk_io_metrics
    datetimes: List[str] = field(default_factory=list)


@dataclass
class LocksMetrics:
    # innodb_trx_lock_count: MetricData
    metadata_lock_count: MetricData
    graphs: List[str]
    tab_name: str = "locks"
    metric_source: MetricSource = MetricSource.none
    datetimes: List[str] = field(default_factory=list)


@dataclass
class MetricInstances:
    dml: DMLMetrics
    replication_lag: ReplicationLagMetrics
    checkpoint: CheckpointMetrics
    buffer_pool_requests: BufferPoolRequestsMetrics
    adaptive_hash_index: AdaptiveHashIndexMetrics
    adaptive_hash_index_hit_ratio: AdaptiveHashIndexHitRatioMetrics
    redo_log: RedoLogMetrics
    redo_log_active_count: RedoLogActiveCountMetrics
    table_cache: TableCacheMetrics
    threads: ThreadMetrics
    temporary_objects: TemporaryObjectMetrics
    aborted_connections: AbortedConnectionsMetrics
    disk_io: DiskIOMetrics
    locks: LocksMetrics


class MetricManager:
    def __init__(self):
        self.reset()

    def reset(self):
        self.worker_start_time: datetime = None
        self.polling_latency: float = None
        self.global_variables: Dict[str, Union[int, str]] = None
        self.global_status: Dict[str, int] = None
        self.replication_lag: int = None
        self.redo_log_size: int = 0

        self.metrics = MetricInstances(
            dml=DMLMetrics(
                graphs=["graph_dml"],
                Queries=MetricData(label="Queries", color=MetricColor.gray, visible=False),
                Com_select=MetricData(label="SELECT", color=MetricColor.blue),
                Com_insert=MetricData(label="INSERT", color=MetricColor.green),
                Com_update=MetricData(label="UPDATE", color=MetricColor.yellow),
                Com_delete=MetricData(label="DELETE", color=MetricColor.red),
                Com_replace=MetricData(
                    label="REPLACE", color=MetricColor.red, visible=False, save_history=False, graphable=False
                ),
                Com_commit=MetricData(
                    label="COMMIT", color=MetricColor.green, visible=False, save_history=True, graphable=False
                ),
                Com_rollback=MetricData(
                    label="ROLLBACK", color=MetricColor.red, visible=False, save_history=False, graphable=False
                ),
            ),
            replication_lag=ReplicationLagMetrics(
                graphs=["graph_replication_lag"],
                lag=MetricData(label="Lag", color=MetricColor.blue, per_second_calculation=False),
            ),
            checkpoint=CheckpointMetrics(
                graphs=["graph_checkpoint"],
                Innodb_checkpoint_age=MetricData(
                    label="Uncheckpointed", color=MetricColor.blue, per_second_calculation=False
                ),
            ),
            buffer_pool_requests=BufferPoolRequestsMetrics(
                graphs=["graph_buffer_pool_requests"],
                Innodb_buffer_pool_read_requests=MetricData(label="Read Requests", color=MetricColor.blue),
                Innodb_buffer_pool_write_requests=MetricData(label="Write Requests", color=MetricColor.green),
                Innodb_buffer_pool_reads=MetricData(label="Disk Reads", color=MetricColor.red),
            ),
            adaptive_hash_index=AdaptiveHashIndexMetrics(
                graphs=["graph_adaptive_hash_index"],
                adaptive_hash_searches=MetricData(label="Hit", color=MetricColor.green),
                adaptive_hash_searches_btree=MetricData(label="Miss", color=MetricColor.red),
            ),
            adaptive_hash_index_hit_ratio=AdaptiveHashIndexHitRatioMetrics(
                graphs=["graph_adaptive_hash_index_hit_ratio"],
                hit_ratio=MetricData(label="Hit Ratio", color=MetricColor.green, per_second_calculation=False),
            ),
            redo_log=RedoLogMetrics(
                graphs=["graph_redo_log", "graph_redo_log_bar"],
                Innodb_lsn_current=MetricData(label="Data Written", color=MetricColor.blue),
            ),
            redo_log_active_count=RedoLogActiveCountMetrics(
                graphs=["graph_redo_log_active_count"],
                Active_redo_log_count=MetricData(
                    label="Active Count", color=MetricColor.blue, per_second_calculation=False, visible=False
                ),
            ),
            table_cache=TableCacheMetrics(
                graphs=["graph_table_cache"],
                Table_open_cache_hits=MetricData(label="Hit", color=MetricColor.green),
                Table_open_cache_misses=MetricData(label="Miss", color=MetricColor.red),
                Table_open_cache_overflows=MetricData(label="Overflow", color=MetricColor.yellow),
            ),
            threads=ThreadMetrics(
                graphs=["graph_threads"],
                Threads_connected=MetricData(label="Connected", color=MetricColor.green, per_second_calculation=False),
                Threads_running=MetricData(label="Running", color=MetricColor.blue, per_second_calculation=False),
            ),
            temporary_objects=TemporaryObjectMetrics(
                graphs=["graph_temporary_objects"],
                Created_tmp_tables=MetricData(label="Tables", color=MetricColor.blue),
                Created_tmp_disk_tables=MetricData(label="Disk", color=MetricColor.red),
                Created_tmp_files=MetricData(label="Files", color=MetricColor.yellow),
            ),
            aborted_connections=AbortedConnectionsMetrics(
                graphs=["graph_aborted_connections"],
                Aborted_clients=MetricData(label="Client (timeout)", color=MetricColor.blue),
                Aborted_connects=MetricData(label="Connects (attempt)", color=MetricColor.red),
            ),
            disk_io=DiskIOMetrics(
                graphs=["graph_disk_io"],
                io_read=MetricData(label="Read", color=MetricColor.blue),
                io_write=MetricData(label="Write", color=MetricColor.green),
            ),
            locks=LocksMetrics(
                graphs=["graph_locks"],
                # innodb_trx_lock_count=MetricData(
                #     label="InnoDB TRX", color=MetricColor.blue, per_second_calculation=False
                # ),
                metadata_lock_count=MetricData(label="Metadata", color=MetricColor.red, per_second_calculation=False),
            ),
        )

    def refresh_data(
        self,
        worker_start_time: datetime,
        polling_latency: float,
        global_variables: Dict[str, Union[int, str]],
        global_status: Dict[str, int],
        innodb_metrics: Dict[str, int],
        disk_io_metrics: Dict[str, int],
        # innodb_trx_lock_metrics: Dict[str, int],
        metadata_lock_metrics: Dict[str, int],
        replication_status: Dict[str, Union[int, str]],
        replication_lag: int,  # this can be from SHOW SLAVE Status/heartbeat table
    ):
        self.worker_start_time = worker_start_time
        self.polling_latency = polling_latency
        self.global_variables = global_variables
        self.global_status = global_status
        self.innodb_metrics = innodb_metrics
        self.disk_io_metrics = disk_io_metrics
        # self.innodb_trx_lock_metrics = innodb_trx_lock_metrics
        self.metadata_lock_metrics = metadata_lock_metrics
        self.replication_status = replication_status
        self.replication_lag = replication_lag

        # Support MySQL 8.0.30+ redo log size variable
        innodb_redo_log_capacity = self.global_variables.get("innodb_redo_log_capacity", 0)
        innodb_log_file_size = round(
            self.global_variables.get("innodb_log_file_size", 0)
            * self.global_variables.get("innodb_log_files_in_group", 1)
        )
        self.redo_log_size = max(innodb_redo_log_capacity, innodb_log_file_size)

        self.update_metrics_per_second_values()

        self.update_metrics_replication_lag()
        self.update_metrics_checkpoint()
        self.update_metrics_adaptive_hash_index_hit_ratio()
        self.update_metrics_locks()

        self.update_metrics_last_value()

        self.metrics.redo_log.redo_log_size = self.redo_log_size

    def add_metric(self, metric_data: MetricData, value: int):
        if metric_data.save_history:
            metric_data.values.append(value)
        else:
            metric_data.values = [value]

    def update_metrics_per_second_values(self):
        for metric_instance in self.metrics.__dict__.values():
            added = False

            metric_source = None  # Initialize as None

            if metric_instance.metric_source == MetricSource.global_status:
                metric_source = self.global_status
            elif metric_instance.metric_source == MetricSource.innodb_metrics:
                metric_source = self.innodb_metrics
            elif metric_instance.metric_source == MetricSource.disk_io_metrics:
                metric_source = self.disk_io_metrics

            if metric_source is None:
                continue  # Skip if there's no metric source

            for metric_name, metric_data in metric_instance.__dict__.items():
                if isinstance(metric_data, MetricData):
                    if metric_data.last_value is None:
                        metric_data.last_value = metric_source.get(metric_name, 0)
                    else:
                        if metric_data.per_second_calculation:
                            metric_status_per_sec = self.get_metric_calculate_per_sec(
                                metric_name, metric_source, format=False
                            )
                        else:
                            metric_status_per_sec = metric_source.get(metric_name, 0)

                        self.add_metric(metric_data, metric_status_per_sec)
                        added = True

            if added:
                metric_instance.datetimes.append(self.worker_start_time.strftime("%d/%m/%y %H:%M:%S"))

    def update_metrics_replication_lag(self):
        if self.replication_status:
            metric_instance = self.metrics.replication_lag
            self.add_metric(metric_instance.lag, self.replication_lag)
            metric_instance.datetimes.append(self.worker_start_time.strftime("%d/%m/%y %H:%M:%S"))

    def update_metrics_adaptive_hash_index_hit_ratio(self):
        hit_ratio = self.get_metric_adaptive_hash_index(format=False)

        if hit_ratio:
            metric_instance = self.metrics.adaptive_hash_index_hit_ratio
            self.add_metric(metric_instance.hit_ratio, hit_ratio)
            metric_instance.datetimes.append(self.worker_start_time.strftime("%d/%m/%y %H:%M:%S"))

    def update_metrics_checkpoint(self):
        (max_checkpoint_age_bytes, checkpoint_age_sync_flush_bytes, _) = self.get_metric_checkpoint_age(format=False)

        metric_instance = self.metrics.checkpoint
        metric_instance.checkpoint_age_max = max_checkpoint_age_bytes
        metric_instance.checkpoint_age_sync_flush = checkpoint_age_sync_flush_bytes

    def update_metrics_locks(self):
        metric_instance = self.metrics.locks
        # self.add_metric(metric_instance.innodb_trx_lock_count, len(self.innodb_trx_lock_metrics))
        self.add_metric(metric_instance.metadata_lock_count, len(self.metadata_lock_metrics))
        metric_instance.datetimes.append(self.worker_start_time.strftime("%d/%m/%y %H:%M:%S"))

    def get_metric_calculate_per_sec(self, metric_name, metric_source=None, format=True):
        if not metric_source:
            metric_source = self.global_status

        for metric_instance in self.metrics.__dict__.values():
            if hasattr(metric_instance, metric_name):
                metric_data: MetricData = getattr(metric_instance, metric_name)

                metric_diff = metric_source.get(metric_name, 0) - metric_data.last_value
                metric_per_sec = round(metric_diff / self.polling_latency)

                if format:
                    return format_number(metric_per_sec)
                else:
                    return metric_per_sec

    def get_metric_checkpoint_age(self, format):
        checkpoint_age_bytes = round(self.global_status.get("Innodb_checkpoint_age", 0))
        max_checkpoint_age_bytes = self.redo_log_size

        if checkpoint_age_bytes + max_checkpoint_age_bytes == 0:
            return "N/A"

        checkpoint_age_sync_flush_bytes = round(max_checkpoint_age_bytes * 0.825)
        checkpoint_age_ratio = round(checkpoint_age_bytes / max_checkpoint_age_bytes * 100, 2)

        if format:
            if checkpoint_age_ratio >= 80:
                color_code = "red"
            elif checkpoint_age_ratio >= 60:
                color_code = "yellow"
            else:
                color_code = "green"

            return f"[{color_code}]{checkpoint_age_ratio}%"
        else:
            return max_checkpoint_age_bytes, checkpoint_age_sync_flush_bytes, checkpoint_age_bytes

    def get_metric_adaptive_hash_index(self, format=True):
        if self.global_variables.get("innodb_adaptive_hash_index") == "OFF":
            return "OFF" if format else None
        elif format:
            smoothed_hit_ratio = self.metrics.adaptive_hash_index_hit_ratio.smoothed_hit_ratio
            if smoothed_hit_ratio is None:
                return "N/A"
            else:
                if smoothed_hit_ratio <= 0.01:
                    return "Inactive"

                color_code = "green" if smoothed_hit_ratio > 70 else "yellow" if smoothed_hit_ratio > 50 else "red"

                return f"[{color_code}]{smoothed_hit_ratio:.2f}%[/{color_code}]"

        current_hits = self.innodb_metrics.get("adaptive_hash_searches", 0)
        current_misses = self.innodb_metrics.get("adaptive_hash_searches_btree", 0)

        hits = current_hits - self.metrics.adaptive_hash_index.adaptive_hash_searches.last_value
        misses = current_misses - self.metrics.adaptive_hash_index.adaptive_hash_searches_btree.last_value
        total_hits_misses = hits + misses
        if total_hits_misses <= 0:
            return None

        hit_ratio = (hits / total_hits_misses) * 100

        smoothing_factor = 0.5
        smoothed_hit_ratio = self.metrics.adaptive_hash_index_hit_ratio.smoothed_hit_ratio

        if smoothed_hit_ratio is None:
            smoothed_hit_ratio = hit_ratio
        else:
            smoothed_hit_ratio = (1 - smoothing_factor) * smoothed_hit_ratio + smoothing_factor * hit_ratio
        self.metrics.adaptive_hash_index_hit_ratio.smoothed_hit_ratio = smoothed_hit_ratio

        return smoothed_hit_ratio

    def update_metrics_last_value(self):
        # We set the last value for specific metrics that need it so they can get per second values
        for metric_instance in self.metrics.__dict__.values():
            if metric_instance.metric_source == MetricSource.global_status:
                metrics_data = self.global_status
            elif metric_instance.metric_source == MetricSource.innodb_metrics:
                metrics_data = self.innodb_metrics
            elif metric_instance.metric_source == MetricSource.disk_io_metrics:
                metrics_data = self.disk_io_metrics

            for metric_name, metric_data in metric_instance.__dict__.items():
                if isinstance(metric_data, MetricData) and metric_data.per_second_calculation:
                    metric_data.last_value = metrics_data.get(metric_name, 0)



from textual.app import ComposeResult
from textual.containers import Container
from textual.reactive import reactive
from textual.widgets import Label


class TopBar(Container):
    host = reactive("", init=False, always_update=True)

    def __init__(
        self, connection_status="", app_version="", host="", help="press [b highlight]q[/b highlight] to return"
    ):
        super().__init__()

        self.topbar_title = Label(
            f" [b light_blue]Dolphie[/b light_blue] :dolphin: [light_blue]v{app_version}", id="topbar_title"
        )
        self.topbar_host = Label(self.host, id="topbar_host")
        self.topbar_help = Label(help, id="topbar_help")

        self.connection_status = connection_status

        self.host = host if host is not None else ""

    def watch_host(self):
        if self.connection_status:
            self.topbar_host.update(f"[[white]{self.connection_status}[/white]] {self.host}")
        else:
            self.topbar_host.update("")

    def compose(self) -> ComposeResult:
        yield self.topbar_title
        yield self.topbar_host
        yield self.topbar_help



import time
from ssl import SSLError

import pymysql
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.Queries import MySQLQueries
from textual.app import App


class Database:
    def __init__(
        self,
        app: App,
        host: str,
        user: str,
        password: str,
        socket: str,
        port: int,
        ssl: str,
        save_connection_id: bool = True,
        auto_connect: bool = True,
    ):
        self.connection: pymysql.Connection = None
        self.connection_id: int = None

        self.app = app
        self.host = host
        self.user = user
        self.password = password
        self.socket = socket
        self.port = port
        self.ssl = ssl
        self.save_connection_id = save_connection_id

        self.max_reconnect_attempts: int = 3
        self.running_query: bool = False
        self.using_ssl: str = None

        if auto_connect:
            self.connect()

    def connect(self):
        try:
            self.connection = pymysql.connect(
                host=self.host,
                user=self.user,
                passwd=self.password,
                unix_socket=self.socket,
                port=int(self.port),
                use_unicode=False,
                ssl=self.ssl,
                autocommit=True,
                connect_timeout=5,
                program_name="Dolphie",
            )
            self.cursor = self.connection.cursor(pymysql.cursors.DictCursor)

            # Get connection ID for processlist filtering
            if self.save_connection_id:
                self.connection_id = self.fetch_value_from_field("SELECT CONNECTION_ID()")

            # Determine if SSL is being used
            self.using_ssl = "ON" if self.fetch_value_from_field("SHOW STATUS LIKE 'Ssl_cipher'", "Value") else "OFF"
        except pymysql.Error as e:
            if len(e.args) == 1:
                raise ManualException(e)
            else:
                raise ManualException(e.args[1])
        except FileNotFoundError:  # Catch SSL file path errors
            raise ManualException("SSL certificate file path isn't valid!")
        except SSLError as e:
            raise ManualException(f"SSL error: {e}")

    def close(self):
        if self.is_connected():
            self.connection.close()

    def is_connected(self) -> bool:
        if self.connection:
            return self.connection.open

        return False

    def execute(self, query, values=None, ignore_error=False):
        if not self.is_connected():
            return None

        if self.running_query:
            self.app.notify(
                "Another query is already running, please repeat action",
                title="Unable to run multiple queries at the same time",
                severity="error",
                timeout=10,
            )
            return None

        error_message = None

        # Prefix all queries with dolphie so they can be identified in the processlist from other people
        query = "/* dolphie */ " + query

        for _ in range(self.max_reconnect_attempts):
            self.running_query = True

            try:
                rows = self.cursor.execute(query, values)
                self.running_query = False

                return rows
            except AttributeError:
                # If the cursor is not defined, reconnect and try again
                self.running_query = False

                self.close()
                self.connect()

                time.sleep(1)
            except pymysql.Error as e:
                self.running_query = False

                if ignore_error:
                    return None
                else:
                    if len(e.args) == 1:
                        error_code = e.args[0]
                    else:
                        error_code = e.args[0]
                        error_message = e.args[1]

                    # Check if the error is due to a connection issue
                    if error_code in (0, 2006, 2013, 2055):
                        # 0: Not connected to MySQL
                        # 2006: MySQL server has gone away
                        # 2013: Lost connection to MySQL server during query
                        # 2055: Lost connection to MySQL server at hostname

                        self.app.notify(
                            f"[b light_blue]{self.host}:{self.port}[/b light_blue]: {error_message}",
                            title="MySQL Connection Lost",
                            severity="error",
                            timeout=10,
                        )

                        self.close()
                        self.connect()

                        self.app.notify(
                            f"[b light_blue]{self.host}:{self.port}[/b light_blue]: Successfully reconnected",
                            title="MySQL Connection Created",
                            severity="success",
                            timeout=10,
                        )

                        time.sleep(1)
                    else:
                        raise ManualException(error_message, query=query)

        if error_message is not None:
            raise ManualException(
                f"{self.host}:{self.port}: Failed to execute query after"
                f" {self.max_reconnect_attempts} reconnection attempts - error: {error_message}",
                query=query,
            )

    def _process_row(self, row):
        processed_row = {}

        for field, value in row.items():
            if isinstance(value, (bytes, bytearray)):
                processed_row[field] = value.decode()
            else:
                processed_row[field] = value

        return processed_row

    def fetchall(self):
        if not self.is_connected():
            return None

        rows = [self._process_row(row) for row in self.cursor.fetchall()]

        if not rows:
            return []

        return rows

    def fetchone(self):
        if not self.is_connected():
            return

        row = self.cursor.fetchone()

        if not row:
            return {}

        return self._process_row(row)

    def fetch_value_from_field(self, query, field=None, values=None):
        if not self.is_connected():
            return None

        self.execute(query, values)
        data = self.cursor.fetchone()

        if not data:
            return None

        field = field or next(iter(data))  # Use field if provided, otherwise get first field
        value = data.get(field)

        if isinstance(value, (bytes, bytearray)):
            return value.decode()

        return value

    def fetch_status_and_variables(self, command):
        if not self.is_connected():
            return None

        command_data = {}

        self.execute(getattr(MySQLQueries, command))
        data = self.fetchall()

        if command in {"status", "variables"}:
            for row in data:
                variable = row["Variable_name"]
                value = row["Value"]

                converted_value = int(value) if value.isnumeric() else value

                command_data[variable] = converted_value
        elif command == "innodb_metrics":
            for row in data:
                metric = row["NAME"]
                value = int(row["COUNT"])

                command_data[metric] = value

        return command_data



from typing import Dict

from dolphie.Modules.Functions import format_query, format_time
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Modules.TabManager import Tab
from rich.syntax import Syntax
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    columns = {
        "OBJECT_TYPE": {"name": "Object Type", "width": 13},
        "OBJECT_SCHEMA": {"name": "Object Schema", "width": 13},
        "OBJECT_NAME": {"name": "Object Name", "width": 25},
        "LOCK_TYPE": {"name": "Lock Type", "width": 20},
        "LOCK_STATUS": {"name": "Lock Status", "width": 11},
        "CODE_SOURCE": {"name": "Code Source", "width": 15},
        "THREAD_SOURCE": {"name": "Thread Source", "width": 15},
        "PROCESSLIST_ID": {"name": "Process ID", "width": 13},
        "PROCESSLIST_USER": {"name": "User", "width": 20},
        "PROCESSLIST_TIME": {"name": "Age", "width": 8},
        "PROCESSLIST_INFO": {"name": "Query", "width": None},
    }

    metadata_locks_datatable = tab.metadata_locks_datatable

    if not metadata_locks_datatable.columns:
        for column_key, column_data in columns.items():
            column_name = column_data["name"]
            metadata_locks_datatable.add_column(column_name, key=column_name, width=column_data["width"])

    for lock in dolphie.metadata_locks:
        lock_id = str(lock["id"])
        row_values = []
        row_height = 1

        for column_id, (column_key, column_data) in enumerate(columns.items()):
            column_name = column_data["name"]
            column_value = lock[column_key]

            # Get height of row based on the how many objects are in the OBJECT_NAME field
            if (
                column_key == "OBJECT_NAME"
                and column_value
                and len(column_value) > column_data["width"]
                and "," in column_value
            ):
                # Truncate the object names to the width of the column
                object_names = [object_name[: column_data["width"]] for object_name in column_value.split(",")]
                thread_value = "\n".join(object_names)

                row_height = len(object_names)
            else:
                thread_value = format_value(lock, column_key, column_value)

            if lock_id in metadata_locks_datatable.rows:
                datatable_value = metadata_locks_datatable.get_row(lock_id)[column_id]

                temp_thread_value = thread_value
                temp_datatable_value = datatable_value

                # If the column is the query, we need to compare the code of the Syntax object
                update_width = False
                if column_key == "PROCESSLIST_INFO":
                    update_width = True

                    if isinstance(thread_value, Syntax):
                        temp_thread_value = thread_value.code
                    if isinstance(datatable_value, Syntax):
                        temp_datatable_value = datatable_value.code

                # Update the datatable if values differ
                if temp_thread_value != temp_datatable_value:
                    metadata_locks_datatable.update_cell(lock_id, column_name, thread_value, update_width=update_width)
            else:
                # Create an array of values to append to the datatable
                row_values.append(thread_value)

        # Add a new row to the datatable
        if row_values:
            metadata_locks_datatable.add_row(*row_values, key=lock_id, height=row_height)

    # Find the ids that exist in datatable but not in metadata_locks
    if dolphie.metadata_locks:
        rows_to_remove = set(metadata_locks_datatable.rows.keys()) - {
            str(lock["id"]) for lock in dolphie.metadata_locks
        }
        for id in rows_to_remove:
            metadata_locks_datatable.remove_row(id)
    else:
        if metadata_locks_datatable.row_count:
            metadata_locks_datatable.clear()

    metadata_locks_datatable.sort("Age", reverse=dolphie.sort_by_time_descending)

    tab.metadata_locks_title.update(f"Metadata Locks ([highlight]{metadata_locks_datatable.row_count}[/highlight])")


def fetch_data(tab: Tab) -> Dict[str, str]:
    dolphie = tab.dolphie

    ########################
    # WHERE clause filters #
    ########################
    where_clause = []

    # Filter user
    if dolphie.user_filter:
        where_clause.append("processlist_user = '%s'" % dolphie.user_filter)

    # Filter database
    if dolphie.db_filter:
        where_clause.append("processlist_db = '%s'" % dolphie.db_filter)

    # Filter hostname/IP
    if dolphie.host_filter:
        # Have to use LIKE since there's a port at the end
        where_clause.append("processlist_host LIKE '%s%%'" % dolphie.host_filter)

    # Filter time
    if dolphie.query_time_filter:
        where_clause.append("processlist_time >= '%s'" % dolphie.query_time_filter)

    # Filter query
    if dolphie.query_filter:
        where_clause.append("(processlist_info LIKE '%%%s%%')" % (dolphie.query_filter))

    if where_clause:
        # Add in our dynamic WHERE clause for filtering
        query = MySQLQueries.metadata_locks.replace("$1", "AND " + " AND ".join(where_clause))
    else:
        query = MySQLQueries.metadata_locks.replace("$1", "")

    dolphie.main_db_connection.execute(query)
    threads = dolphie.main_db_connection.fetchall()

    return threads


def format_value(lock: dict, column_key: str, value: str) -> str:
    formatted_value = value

    # OBJECT_NAME is in the format "schema/table" sometimes where OBJECT_SCHEMA is empty,
    # so I want to split OBJECT_NAME and correct it if necessary
    if column_key == "OBJECT_SCHEMA" and not value and lock["OBJECT_NAME"] and "/" in lock["OBJECT_NAME"]:
        formatted_value = lock["OBJECT_NAME"].split("/")[0]
    elif column_key == "OBJECT_NAME" and value and "/" in value:
        formatted_value = value.split("/")[1]
    elif value is None or value == "":
        formatted_value = "[dark_gray]N/A"
    elif column_key == "PROCESSLIST_INFO":
        formatted_value = format_query(value)
    elif column_key == "LOCK_STATUS":
        if value == "GRANTED":
            formatted_value = f"[green]{value}[/green]"
        elif value == "PENDING":
            formatted_value = f"[red]{value}[/red]"
    elif column_key == "LOCK_TYPE":
        if value == "EXCLUSIVE":
            formatted_value = f"[yellow]{value}[/yellow]"
    elif column_key == "PROCESSLIST_TIME":
        formatted_value = format_time(value)
    elif column_key == "CODE_SOURCE":
        formatted_value = value.split(":")[0]
    elif column_key == "THREAD_SOURCE":
        formatted_value = value.split("/")[-1]

        if formatted_value == "one_connection":
            formatted_value = "user_connection"

    return formatted_value



from dolphie.Modules.ManualException import ManualException
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    Checkbox,
    Input,
    Label,
    RadioButton,
    RadioSet,
    Select,
    Static,
)
from textual_autocomplete import AutoComplete, Dropdown, DropdownItem


class HostSetupModal(ModalScreen):
    CSS = """
        HostSetupModal > Vertical {
            background: #131626;
            border: tall #384673;
            height: auto;
            width: 70;
        }
        HostSetupModal > Vertical > * {
            width: auto;
            height: auto;
            align: center middle;
        }
        HostSetupModal > Vertical > Vertical  > Container {
            width: 100%;
            height: auto;
        }
        HostSetupModal Label {
            text-style: bold;
            width: 100%;
            content-align: center middle;
            padding-bottom: 1;
        }
        HostSetupModal Input {
            width: 100%;
            content-align: center middle;
        }
        HostSetupModal .main_container {
            width: 100%;
            content-align: center middle;
        }
        HostSetupModal RadioSet {
            background: #131626;
            border: none;
            padding-bottom: 1;
            align: center middle;
            width: 100%;
            layout: horizontal;
        }
        HostSetupModal AutoComplete {
            width: 100%;
            height: auto;
        }
        HostSetupModal #password {
            width: 84%;
        }
        HostSetupModal #show_password {
            max-width: 12%;
            background:  #262c4b;
            border: blank #344063;
        }
        HostSetupModal #show_password:hover {
            background:  #313960;
            border: blank #344063;
        }
        HostSetupModal #show_password:focus {
            background:  #313960;
            border: blank #344063;
        }
        HostSetupModal #modal_footer {
            color: #d3565c;
            width: 100%;
            padding-bottom: 0;
            padding-top: 1;
            margin: 0 2;
        }
        HostSetupModal Select {
            width: 100%;
            margin: 0 2;
            margin-bottom: 1;
        }
        HostSetupModal SelectCurrent Static#label {
            color: #606e88;
        }
        HostSetupModal SelectCurrent.-has-value Static#label {
            color: #e9e9e9;
        }
        HostSetupModal Select:focus > SelectCurrent {
            border: tall #384673;
        }
        HostSetupModal SelectCurrent {
            background: #111322;
            border: tall #252e49;
        }
        HostSetupModal Select > OptionList:focus {
            margin: 0 0 0 0;
            height: auto;
            max-height: 15;
            border: tall #3c476b;
        }
        HostSetupModal OptionList {
            background: #111322;
            border: tall #252e49;
            width: 100%;
            height: 15;
            margin: 0 1 0 1;
        }
        HostSetupModal OptionList:focus {
            border: tall #475484;
        }
        HostSetupModal OptionList > .option-list--option-highlighted {
            text-style: none;
            background: #131626;
        }
        HostSetupModal OptionList:focus > .option-list--option-highlighted {
            background: #283048;
        }
        HostSetupModal OptionList > .option-list--option-hover {
            background: #283048;
        }
        HostSetupModal OptionList > .option-list--option-hover-highlighted {
            background: #283048;
            text-style: none;
        }
        HostSetupModal OptionList:focus > .option-list--option-hover-highlighted {
            background: #283048;
            text-style: none;
        }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]

    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        ssl: dict,
        socket_file: str,
        available_hosts: list,
        hostgroups: dict,
        error_message: ManualException = None,
    ):
        super().__init__()

        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.ssl = ssl
        self.socket_file = socket_file

        if self.host and self.port:
            self.host = f"{self.host}:{self.port}"

        self.dropdown_items = []
        if available_hosts:
            self.dropdown_items = [DropdownItem(id) for id in sorted(available_hosts)]

        self.hostgroups = []
        if hostgroups:
            self.hostgroups = [(host, host) for host in hostgroups]

        self.error_message = error_message

    def on_mount(self) -> None:
        footer = self.query_one("#modal_footer", Static)

        footer.display = False
        if self.error_message:
            footer.update(self.error_message.output())
            footer.display = True

        if self.ssl:
            self.query_one("#ssl", Checkbox).value = True

            if self.ssl.get("required"):
                ssl_mode = "REQUIRED"
            elif self.ssl.get("check_hostname") is False:
                ssl_mode = "VERIFY_CA"
            elif self.ssl.get("check_hostname") is True:
                ssl_mode = "VERIFY_IDENTITY"

            self.query_one(f"#{ssl_mode}", RadioButton).value = True
            self.query_one("#ssl_ca", Input).value = self.ssl.get("ca", "")
            self.query_one("#ssl_cert", Input).value = self.ssl.get("cert", "")
            self.query_one("#ssl_key", Input).value = self.ssl.get("key", "")
        else:
            self.query_one("#container_ssl", Container).display = False

        if self.socket_file:
            self.query_one("#socket_file", Input).value = self.socket_file

    def compose(self) -> ComposeResult:
        with Vertical():
            with Vertical(classes="main_container"):
                yield Label("Host Setup")
                yield Select(
                    options=self.hostgroups,
                    id="hostgroup",
                    prompt="Select a hostgroup (optional)",
                )
                yield AutoComplete(
                    Input(
                        value=self.host,
                        id="host",
                        placeholder="Start typing to search for a host - format is host:port",
                    ),
                    Dropdown(id="dropdown_items", items=self.dropdown_items),
                )

                yield Input(id="username", value=self.username, placeholder="Username")
                with Horizontal():
                    yield Input(id="password", value=self.password, placeholder="Password", password=True)
                    yield Button("Show", id="show_password")
                yield Input(id="socket_file", placeholder="Socket file (optional)")
                yield Checkbox("Enable SSL", id="ssl")
                with Container(id="container_ssl"):
                    yield RadioSet(
                        *(
                            [
                                RadioButton(ssl_mode, id=ssl_mode)
                                for ssl_mode in ["REQUIRED", "VERIFY_CA", "VERIFY_IDENTITY"]
                            ]
                        ),
                        id="ssl_mode",
                    )
                    yield Input(id="ssl_ca", placeholder="CA File (optional)")
                    yield Input(id="ssl_cert", placeholder="Client Certificate File (optional)")
                    yield Input(id="ssl_key", placeholder="Client Key File (optional)")
            with Horizontal(classes="button_container"):
                yield Button("Submit", id="submit", variant="primary")
                yield Button("Cancel", id="cancel")
            yield Label(id="modal_footer")

    @on(Select.Changed, "#hostgroup")
    def hostgroup_changed(self, event: Select.Changed):
        if event.value == Select.BLANK:
            self.query_one("#host", Input).disabled = False
            self.query_one("#username", Input).disabled = False
            self.query_one("#password", Input).disabled = False
            self.query_one("#socket_file", Input).disabled = False
            self.query_one("#ssl", Checkbox).disabled = False
            self.query_one("#show_password", Button).disabled = False
        elif event.value:
            self.query_one("#host", Input).disabled = True
            self.query_one("#username", Input).disabled = True
            self.query_one("#password", Input).disabled = True
            self.query_one("#socket_file", Input).disabled = True
            self.query_one("#ssl", Checkbox).value = False
            self.query_one("#ssl", Checkbox).disabled = True
            self.query_one("#show_password", Button).disabled = True

    @on(RadioSet.Changed, "#ssl_mode")
    def ssl_mode_changed(self, event: RadioSet.Changed):
        if event.pressed.id in ["VERIFY_CA", "VERIFY_IDENTITY"]:
            self.query_one("#ssl_ca", Input).placeholder = "CA File"
            self.query_one("#ssl_cert", Input).placeholder = "Client Certificate File (optional)"
            self.query_one("#ssl_key", Input).placeholder = "Client Key File (optional)"
        else:
            self.query_one("#ssl_ca", Input).placeholder = "CA File (optional)"
            self.query_one("#ssl_cert", Input).placeholder = "Client Certificate File (optional)"
            self.query_one("#ssl_key", Input).placeholder = "Client Key File (optional)"

    @on(Checkbox.Changed, "#ssl")
    def ssl_changed(self, event: Checkbox.Changed):
        if event.value:
            self.query_one("#container_ssl", Container).display = True
        else:
            self.query_one("#container_ssl", Container).display = False

    @on(Button.Pressed, "#show_password")
    def on_show_password_pressed(self, event: Button.Pressed) -> None:
        password = self.query_one("#password", Input)
        show_password_button = self.query_one("#show_password", Button)

        if password.password:
            show_password_button.label = "Hide"
        else:
            show_password_button.label = "Show"

        password.password = not password.password

    @on(Button.Pressed, "#submit")
    def on_submit_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "submit":
            error_message = None

            host = self.query_one("#host", Input)
            hostgroup = self.query_one("#hostgroup", Select)
            username = self.query_one("#username", Input)
            password = self.query_one("#password", Input)

            ssl_mode = (
                self.query_one("#ssl_mode", RadioSet).pressed_button.id
                if self.query_one("#ssl_mode", RadioSet).pressed_button
                else None
            )
            ssl_ca = self.query_one("#ssl_ca", Input).value
            ssl_cert = self.query_one("#ssl_cert", Input).value
            ssl_key = self.query_one("#ssl_key", Input).value

            ssl = {}
            if self.query_one("#ssl", Checkbox).value and ssl_mode:
                if ssl_mode == "REQUIRED":
                    ssl["required"] = True
                elif ssl_mode == "VERIFY_CA":
                    if not ssl_ca:
                        error_message = "SSL mode VERIFY_CA requires CA File to be specified"

                    ssl["check_hostname"] = False
                    ssl["verify_mode"] = True
                elif ssl_mode == "VERIFY_IDENTITY":
                    if not ssl_ca:
                        error_message = "SSL mode VERIFY_IDENTITY requires CA File to be specified"

                    ssl["check_hostname"] = True
                    ssl["verify_mode"] = True

                if ssl_ca:
                    ssl["ca"] = ssl_ca
                if ssl_cert:
                    ssl["cert"] = ssl_cert
                if ssl_key:
                    ssl["key"] = ssl_key

            socket_file = self.query_one("#socket_file", Input)

            hostgroup_value = hostgroup.value
            if hostgroup_value == Select.BLANK:
                hostgroup_value = None

            modal_footer = self.query_one("#modal_footer", Label)
            host_split = host.value.split(":")
            if len(host_split) == 2:
                try:
                    int(host_split[1])
                except ValueError:
                    error_message = "Port must be a valid integer"
            elif len(host_split) > 2:
                error_message = "Host cannot contain more than one colon"

            if not host.value and not hostgroup_value:
                error_message = "You must specify either a host or hostgroup"

            if error_message:
                modal_footer.update(error_message)
                modal_footer.display = True
                return

            self.dismiss(
                {
                    "host": host.value,
                    "hostgroup": hostgroup_value,
                    "username": username.value,
                    "password": password.value,
                    "ssl": ssl,
                    "socket_file": socket_file.value,
                }
            )
        else:
            self.app.pop_screen()

    @on(Input.Submitted, "#password")
    def on_input_submitted(self):
        self.query_one("#submit", Button).press()

    @on(Button.Pressed, "#cancel")
    def on_cancel_pressed(self, event: Button.Pressed) -> None:
        self.app.pop_screen()



from dolphie.Widgets.topbar import TopBar
from textual import events
from textual.containers import Center
from textual.screen import Screen
from textual.widgets import Static


class CommandScreen(Screen):
    CSS = """
        CommandScreen Center {
            padding: 1;
        }
        CommandScreen Center > Static {
            padding-left: 1;
            padding-right: 1;
            background: #101626;
            border: tall #1d253e;
            width: auto;
        }
    """

    def __init__(self, connection_status, app_version, host, data):
        super().__init__()
        self.connection_status = connection_status
        self.app_version = app_version
        self.host = host
        self.data = data

    def on_key(self, event: events.Key):
        if event.key == "q" and self.screen.is_attached:
            self.app.pop_screen()

    def compose(self):
        yield TopBar(connection_status=self.connection_status, app_version=self.app_version, host=self.host)
        yield Center(Static(self.data, shrink=True))



from typing import Dict

from dolphie.DataTypes import ProcesslistThread
from dolphie.Modules.Functions import format_number
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Modules.TabManager import Tab
from rich.syntax import Syntax
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    if not dolphie.performance_schema_enabled and dolphie.use_performance_schema:
        dolphie.notify("Performance Schema is not enabled on this host, using Information Schema instead")
        dolphie.use_performance_schema = False

    columns = [
        {"name": "Process ID", "field": "id", "width": 11, "format_number": False},
        {"name": "Protocol", "field": "protocol", "width": 8, "format_number": False},
        {"name": "Username", "field": "user", "width": 20, "format_number": False},
    ]

    if dolphie.show_additional_query_columns:
        columns.extend(
            [
                {"name": "Hostname/IP", "field": "host", "width": 16, "format_number": False},
                {"name": "Database", "field": "db", "width": 13, "format_number": False},
            ]
        )

    columns.extend(
        [
            {"name": "Command", "field": "command", "width": 8, "format_number": False},
            {"name": "State", "field": "state", "width": 20, "format_number": False},
            {"name": "TRX State", "field": "trx_state", "width": 9, "format_number": False},
            {"name": "R-Lock", "field": "trx_rows_locked", "width": 7, "format_number": True},
            {"name": "R-Mod", "field": "trx_rows_modified", "width": 7, "format_number": True},
        ]
    )

    if dolphie.show_additional_query_columns and dolphie.global_variables.get("innodb_thread_concurrency"):
        columns.append({"name": "Tickets", "field": "trx_concurrency_tickets", "width": 8, "format_number": True})

    if dolphie.show_trxs_only:
        columns.append(
            {"name": "TRX Age", "field": "trx_time", "width": 9, "format_number": False},
        )

    columns.extend(
        [
            {"name": "Age", "field": "formatted_time", "width": 9, "format_number": False},
            {"name": "Query", "field": "formatted_query", "width": None, "format_number": False},
            {"name": "time_seconds", "field": "time", "width": 0, "format_number": False},
        ]
    )

    processlist_datatable = tab.processlist_datatable

    # Clear table if columns change
    if len(processlist_datatable.columns) != len(columns):
        processlist_datatable.clear(columns=True)

    # Add columns to the datatable if it is empty
    if not processlist_datatable.columns:
        for column_data in columns:
            column_name = column_data["name"]
            column_width = column_data["width"]
            processlist_datatable.add_column(column_name, key=column_name, width=column_width)

    # Iterate through processlist_threads
    for thread_id, thread in dolphie.processlist_threads.items():
        row_values = []

        for column_id, (column_data) in enumerate(columns):
            column_name = column_data["name"]
            column_field = column_data["field"]
            column_format_number = column_data["format_number"]
            column_value = getattr(thread, column_field)

            thread_value = format_number(column_value) if column_format_number else column_value
            if thread_id in processlist_datatable.rows:
                datatable_value = processlist_datatable.get_row(thread_id)[column_id]

                # Store the code of the Syntax object if it's a query
                temp_thread_value = thread_value
                temp_datatable_value = datatable_value

                # If the column is the query, we need to compare the code of the Syntax object
                update_width = False
                if column_field == "formatted_query":
                    update_width = True
                    if isinstance(thread_value, Syntax):
                        temp_thread_value = thread_value.code
                    if isinstance(datatable_value, Syntax):
                        temp_datatable_value = datatable_value.code

                # Update the datatable if values differ
                if temp_thread_value != temp_datatable_value or column_field == "formatted_time":
                    processlist_datatable.update_cell(thread_id, column_name, thread_value, update_width=update_width)
            else:
                # Create an array of values to append to the datatable
                row_values.append(thread_value)

        # Add a new row to the datatable
        if row_values:
            processlist_datatable.add_row(*row_values, key=thread_id)

    # Remove rows from processlist_datatable that no longer exist in processlist_threads
    if dolphie.processlist_threads:
        rows_to_remove = set(processlist_datatable.rows.keys()) - set(dolphie.processlist_threads.keys())
        for id in rows_to_remove:
            processlist_datatable.remove_row(id)
    else:
        if processlist_datatable.row_count:
            processlist_datatable.clear()

    processlist_datatable.sort("time_seconds", reverse=dolphie.sort_by_time_descending)

    tab.processlist_title.update(f"Processlist ([highlight]{processlist_datatable.row_count}[/highlight])")


def fetch_data(tab: Tab) -> Dict[str, ProcesslistThread]:
    dolphie = tab.dolphie

    if dolphie.use_performance_schema:
        processlist_query = MySQLQueries.ps_query
    else:
        processlist_query = MySQLQueries.pl_query

    ########################
    # WHERE clause filters #
    ########################
    where_clause = []

    # Filter out idle threads if specified
    if not dolphie.show_idle_threads:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_command != 'Sleep' AND processlist_command NOT LIKE 'Binlog Dump%') AND (processlist_info"
                " IS NOT NULL OR trx_query IS NOT NULL) AND IFNULL(processlist_state, '') NOT LIKE 'Group Replication"
                " Module%'"
            )
        else:
            where_clause.append(
                "(Command != 'Sleep' AND Command NOT LIKE 'Binlog Dump%') AND (Info IS NOT NULL OR trx_query IS NOT"
                " NULL) AND IFNULL(State, '') NOT LIKE 'Group Replication Module%'"
            )

    # Only show running transactions only
    if dolphie.show_trxs_only:
        where_clause.append("trx_state != ''")

    # Filter user
    if dolphie.user_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_user = '%s'" % dolphie.user_filter)
        else:
            where_clause.append("User = '%s'" % dolphie.user_filter)

    # Filter database
    if dolphie.db_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_db = '%s'" % dolphie.db_filter)
        else:
            where_clause.append("db = '%s'" % dolphie.db_filter)

    # Filter hostname/IP
    if dolphie.host_filter:
        # Have to use LIKE since there's a port at the end
        if dolphie.use_performance_schema:
            where_clause.append("processlist_host LIKE '%s%%'" % dolphie.host_filter)
        else:
            where_clause.append("Host LIKE '%s%%'" % dolphie.host_filter)

    # Filter time
    if dolphie.query_time_filter:
        if dolphie.use_performance_schema:
            where_clause.append("processlist_time >= '%s'" % dolphie.query_time_filter)
        else:
            where_clause.append("Time >= '%s'" % dolphie.query_time_filter)

    # Filter query
    if dolphie.query_filter:
        if dolphie.use_performance_schema:
            where_clause.append(
                "(processlist_info LIKE '%%%s%%' OR trx_query LIKE '%%%s%%')"
                % (dolphie.query_filter, dolphie.query_filter),
            )
        else:
            where_clause.append("Info LIKE '%%%s%%'" % dolphie.query_filter)

    # Add in our dynamic WHERE clause for filtering
    if where_clause:
        processlist_query = processlist_query.replace("$1", "AND " + " AND ".join(where_clause))
    else:
        processlist_query = processlist_query.replace("$1", "")

    processlist_threads = {}
    # Run the processlist query
    dolphie.main_db_connection.execute(processlist_query)
    threads = dolphie.main_db_connection.fetchall()

    for thread in threads:
        # Don't include Dolphie's threads
        if (
            dolphie.main_db_connection.connection_id == thread["id"]
            or dolphie.secondary_db_connection.connection_id == thread["id"]
        ):
            continue

        # Use trx_query over Performance Schema query since it's more accurate
        if dolphie.use_performance_schema and thread["trx_query"]:
            thread["query"] = thread["trx_query"]
        else:
            thread["query"] = thread["query"]

        host = thread["host"].split(":")[0]
        thread["host"] = dolphie.get_hostname(host)

        processlist_threads[str(thread["id"])] = ProcesslistThread(thread)

    return processlist_threads



from datetime import datetime, timedelta

from dolphie.Modules.Functions import format_bytes, format_number
from dolphie.Modules.MetricManager import MetricData
from dolphie.Modules.TabManager import Tab
from dolphie.Panels import replication_panel
from rich.style import Style
from rich.table import Table


def create_panel(tab: Tab) -> Table:
    dolphie = tab.dolphie

    global_status = dolphie.global_status
    global_variables = dolphie.global_variables
    binlog_status = dolphie.binlog_status

    table_title_style = Style(color="#bbc8e8", bold=True)

    ################
    # Information #
    ###############
    table_information = Table(show_header=False, box=None, title="Host Information", title_style=table_title_style)

    if dolphie.replicaset:
        host_type = "InnoDB ReplicaSet"
    elif dolphie.innodb_cluster_read_replica:
        host_type = "InnoDB Cluster Read Replica"
    elif dolphie.innodb_cluster:
        host_type = "InnoDB Cluster"
    elif dolphie.group_replication:
        host_type = "Group Replication"
    elif dolphie.galera_cluster:
        host_type = "Galera Cluster"
    else:
        host_type = "MySQL"

    runtime = str(datetime.now() - dolphie.dolphie_start_time).split(".")[0]

    replicas = 0
    if dolphie.replica_manager.available_replicas:
        replicas = len(dolphie.replica_manager.available_replicas)

    table_information.add_column()
    table_information.add_column(min_width=25, max_width=27)
    table_information.add_row("[label]Version", f"{dolphie.host_distro} {dolphie.mysql_version}")
    table_information.add_row(
        "[label]", "%s (%s)" % (global_variables["version_compile_os"], global_variables["version_compile_machine"])
    )
    table_information.add_row("[label]Type", f"{host_type} [label]SSL[/label] {dolphie.main_db_connection.using_ssl}")
    table_information.add_row("[label]Uptime", str(timedelta(seconds=global_status["Uptime"])))
    table_information.add_row("[label]Runtime", f"{runtime} [label]Latency[/label] {dolphie.refresh_latency}s")
    table_information.add_row("[label]Replicas", "%s" % replicas)
    table_information.add_row(
        "[label]Threads",
        "[label]con[/label] %s[highlight]/[/highlight][label]run[/label]"
        " %s[highlight]/[/highlight][label]cac[/label] %s"
        % (
            format_number(global_status["Threads_connected"]),
            format_number(global_status["Threads_running"]),
            format_number(global_status["Threads_cached"]),
        ),
    )
    table_information.add_row(
        "[label]Tables",
        "[label]open[/label] %s[highlight]/[/highlight][label]opened[/label] %s"
        % (
            format_number(global_status["Open_tables"]),
            format_number(global_status["Opened_tables"]),
        ),
    )

    tab.dashboard_host_information.update(table_information)

    ###########
    # InnoDB  #
    ###########
    table_innodb = Table(show_header=False, box=None, title="InnoDB", title_style=table_title_style)

    table_innodb.add_column()
    table_innodb.add_column(width=9)

    history_list_length = dolphie.innodb_metrics.get("trx_rseg_history_len", "N/A")

    # Calculate InnoDB memory read hit efficiency
    ib_pool_disk_reads = global_status.get("Innodb_buffer_pool_reads", 0)
    ib_pool_mem_reads = global_status.get(
        "Innodb_buffer_pool_read_requests", 1
    )  # Default to 1 to avoid division by zero

    if ib_pool_disk_reads >= ib_pool_mem_reads:
        innodb_efficiency = "[red]0.00%"
    else:
        efficiency = 100 - (ib_pool_disk_reads / ib_pool_mem_reads * 100)

        if efficiency > 90:
            color_code = "green"
        elif efficiency > 80:
            color_code = "yellow"
        else:
            color_code = "red"

        innodb_efficiency = f"[{color_code}]{efficiency:.2f}%"

    # Add data to our table
    table_innodb.add_row("[label]Read Hit", innodb_efficiency)
    table_innodb.add_row("[label]Chkpt Age", dolphie.metric_manager.get_metric_checkpoint_age(format=True))
    table_innodb.add_row("[label]AHI Hit", dolphie.metric_manager.get_metric_adaptive_hash_index())

    bp_instances = global_variables.get("innodb_buffer_pool_instances", "N/A")
    if bp_instances != "N/A":
        plural = "s" if bp_instances > 1 else ""
        table_innodb.add_row("[label]BP Instance" + plural, format_number(bp_instances))
    else:
        table_innodb.add_row("[label]BP Instance", bp_instances)

    table_innodb.add_row("[label]BP Size", format_bytes(global_variables["innodb_buffer_pool_size"]))
    table_innodb.add_row(
        "[label]BP Available",
        format_bytes(
            dolphie.global_variables["innodb_buffer_pool_size"] - dolphie.global_status["Innodb_buffer_pool_bytes_data"]
        ),
    )
    table_innodb.add_row("[label]BP Dirty", format_bytes(global_status["Innodb_buffer_pool_bytes_dirty"]))
    table_innodb.add_row("[label]History List", format_number(history_list_length))

    tab.dashboard_innodb.update(table_innodb)

    ##############
    # Binary Log #
    ##############
    table_primary = Table(show_header=False, box=None, title="Binary Log", title_style=table_title_style)

    if not binlog_status:
        table_primary.add_column(justify="center")
        table_primary.add_row("\n\n\n[b][label]Disabled")
    else:
        table_primary.add_column()
        table_primary.add_column(max_width=40)

        if dolphie.previous_binlog_position == 0:
            diff_binlog_position = 0
        elif dolphie.previous_binlog_position > binlog_status["Position"]:
            diff_binlog_position = "Binlog Rotated"
        else:
            diff_binlog_position = format_bytes(binlog_status["Position"] - dolphie.previous_binlog_position)

        binlog_cache = 100
        binlog_cache_disk = global_status["Binlog_cache_disk_use"]
        binlog_cache_mem = global_status["Binlog_cache_use"]
        if binlog_cache_disk and binlog_cache_mem:
            if binlog_cache_disk >= binlog_cache_mem:
                innodb_efficiency = "[red]0.00%"
            elif binlog_cache_mem > binlog_cache_disk:
                binlog_cache = round(100 - (binlog_cache_disk / binlog_cache_mem), 2)

        table_primary.add_row("[label]File name", binlog_status["File"])
        table_primary.add_row(
            "[label]Position",
            "%s" % (str(binlog_status["Position"])),
        )
        table_primary.add_row(
            "[label]Size",
            "%s" % format_bytes(binlog_status["Position"]),
        )
        table_primary.add_row("[label]Diff", str(diff_binlog_position))
        table_primary.add_row("[label]Cache Hit", f"{binlog_cache}%")

        binlog_format = global_variables.get("binlog_format", "N/A")
        binlog_row_image = None
        if binlog_format == "ROW":
            binlog_row_image = global_variables.get("binlog_row_image", "N/A")
            table_primary.add_row("[label]Format", "{} ({})".format(binlog_format, binlog_row_image))
        else:
            table_primary.add_row("[label]Format", binlog_format, binlog_row_image)

        gtid_mode = global_variables.get("gtid_mode", "N/A")
        table_primary.add_row("[label]GTID", gtid_mode)

        binlog_compression = global_variables.get("binlog_transaction_compression", "N/A")
        # binlog_compression_percentage = ""
        # if binlog_compression == "ON":
        #     if dolphie.binlog_transaction_compression_percentage:
        #         binlog_compression_percentage = f" ({dolphie.binlog_transaction_compression_percentage}% gain)"
        #     else:
        #         binlog_compression_percentage = " (N/A gain)"

        table_primary.add_row("[label]Compression", binlog_compression)

        # Save some global_variables to be used in next refresh
        if dolphie.binlog_status:
            dolphie.previous_binlog_position = dolphie.binlog_status["Position"]

    tab.dashboard_binary_log.update(table_primary)

    ###############
    # Replication #
    ###############
    if dolphie.replication_status and not dolphie.panels.replication.visible:
        tab.dashboard_replication.display = True
        tab.dashboard_replication.update(replication_panel.create_replication_table(tab, dashboard_table=True))
    else:
        tab.dashboard_replication.display = False
    ###############
    # Statistics #
    ###############
    table_stats = Table(show_header=False, box=None, title="Statistics/s", title_style=table_title_style)

    table_stats.add_column()
    table_stats.add_column(min_width=6)

    # Add DML statistics
    metrics = dolphie.metric_manager.metrics.dml
    metric_labels = {
        "Queries": "Queries",
        "SELECT": "Com_select",
        "INSERT": "Com_insert",
        "UPDATE": "Com_update",
        "DELETE": "Com_delete",
        "REPLACE": "Com_replace",
        "COMMIT": "Com_commit",
        "ROLLBACK": "Com_rollback",
    }

    for label, metric_name in metric_labels.items():
        metric_data: MetricData = getattr(metrics, metric_name)

        if metric_data.values:
            table_stats.add_row(f"[label]{label}", format_number(metric_data.values[-1]))
        else:
            table_stats.add_row(f"[label]{label}", "0")

    tab.dashboard_statistics.update(table_stats)



import re

from dolphie.DataTypes import HotkeyCommands
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    Checkbox,
    Input,
    Label,
    RadioButton,
    RadioSet,
    Static,
)
from textual_autocomplete import AutoComplete, Dropdown, DropdownItem


class CommandModal(ModalScreen):
    CSS = """
        CommandModal > Vertical {
            background: #131626;
            border: tall #384673;
            height: auto;
            width: auto;
        }
        CommandModal > Vertical > * {
            width: auto;
            height: auto;
        }
        CommandModal #filter_radio_buttons {
            margin-bottom: 1;
        }
        CommandModal #kill_container {
            width: 100%;
            height: 6;
        }
        CommandModal Label {
            text-style: bold;
            width: 100%;
            content-align: center middle;
            padding-bottom: 1;
        }
        CommandModal #error_response {
            color: #fe5c5c;
            width: 100%;
            height: auto;
            content-align: center middle;
            padding-bottom: 1;
        }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]

    def __init__(self, command, message, processlist_data=None, host_cache_data=None):
        super().__init__()
        self.command = command
        self.message = message
        self.processlist_data = processlist_data
        self.host_cache_data = host_cache_data

        self.dropdown_items = []
        if processlist_data:
            sorted_keys = sorted(processlist_data.keys(), key=lambda x: int(x))
            self.dropdown_items = [DropdownItem(thread_id) for thread_id in sorted_keys]

    def compose(self) -> ComposeResult:
        with Vertical():
            with Vertical():
                yield Label(self.message)
                with RadioSet(id="filter_radio_buttons"):
                    yield RadioButton("User", id="user")
                    yield RadioButton("Host/IP", id="host")
                    yield RadioButton("Database", id="db")
                    yield RadioButton("Query Text", id="query_text")
                    yield RadioButton("Query Time", id="query_time")
                with Vertical(id="kill_container"):
                    with RadioSet(id="kill_radio_buttons"):
                        yield RadioButton("Username", id="username")
                        yield RadioButton("Host/IP", id="host")
                        yield RadioButton("Time range", id="time_range")
                    yield Checkbox("Include sleeping queries", id="sleeping_queries")
                yield AutoComplete(
                    Input(id="modal_input"),
                    Dropdown(id="dropdown_items", items=self.dropdown_items),
                )
                yield Static(id="error_response")
            with Horizontal(classes="button_container"):
                yield Button("Submit", id="submit", variant="primary")
                yield Button("Cancel", id="cancel")

    def on_mount(self):
        input = self.query_one("#modal_input", Input)
        filter_radio_buttons = self.query_one("#filter_radio_buttons", RadioSet)
        kill_container = self.query_one("#kill_container", Vertical)
        self.query_one("#error_response", Static).display = False

        filter_radio_buttons.display = False
        kill_container.display = False

        if self.command == HotkeyCommands.thread_filter:
            filter_radio_buttons.display = True
            filter_radio_buttons.focus()

            input.placeholder = "Select an option from above"
        elif self.command == HotkeyCommands.thread_kill_by_parameter:
            sleeping_queries_checkbox = self.query_one("#sleeping_queries", Checkbox)
            sleeping_queries_checkbox.toggle()

            kill_container.display = True

            kill_radio_buttons = self.query_one("#kill_radio_buttons", RadioSet)
            kill_radio_buttons.focus()

            input.placeholder = "Select an option from above"
        elif self.command == HotkeyCommands.rename_tab:
            input.placeholder = "Colors can be added by wrapping them in []"
            input.styles.width = 50
            input.focus()
        elif self.command == HotkeyCommands.variable_search:
            input.placeholder = "Input 'all' to show everything"
            input.focus()
        elif self.command in [HotkeyCommands.show_thread, HotkeyCommands.thread_kill_by_id]:
            input.placeholder = "Input a Process ID"
            input.focus()
        elif self.command == HotkeyCommands.refresh_interval:
            input.placeholder = "Input a refresh interval (seconds)"
            input.focus()
        else:
            input.focus()

    def on_input_submitted(self):
        self.query_one("#submit", Button).press()

    def create_dropdown_items(self, field):
        self.dropdown_items = []

        if field:
            sorted_array = sorted(set(getattr(thread, field) for thread in self.processlist_data.values()))
            self.dropdown_items = [DropdownItem(value) for value in sorted_array]

        self.query_one("#dropdown_items", Dropdown).items = self.dropdown_items

    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        modal_input = self.query_one("#modal_input", Input)

        self.create_dropdown_items(None)  # empty string to clear dropdown items

        if self.command == HotkeyCommands.thread_filter:
            if event.pressed.id == "db":
                self.create_dropdown_items("db")
                modal_input.placeholder = "Database name"
            elif event.pressed.id == "host":
                self.create_dropdown_items("host")
                modal_input.placeholder = "Hostname or IP address"
            elif event.pressed.id == "query_text":
                modal_input.placeholder = "Partial query text"
            elif event.pressed.id == "query_time":
                modal_input.placeholder = "Minimum query time (in seconds)"
            elif event.pressed.id == "user":
                self.create_dropdown_items("user")
                modal_input.placeholder = "Username"
        elif self.command == HotkeyCommands.thread_kill_by_parameter:
            if event.pressed.id == "username":
                self.create_dropdown_items("user")
                modal_input.placeholder = "Username"
            elif event.pressed.id == "host":
                self.create_dropdown_items("host")
                modal_input.placeholder = "Hostname or IP address"
            elif event.pressed.id == "time_range":
                modal_input.placeholder = "Time range (ex. 10-20)"

        modal_input.focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id != "submit":
            self.app.pop_screen()
            return

        modal_input = self.query_one("#modal_input", Input).value
        if not modal_input:
            self.update_error_response("Input cannot be empty")
            return

        if self.command == HotkeyCommands.thread_filter:
            filter_options = {
                "user": "User",
                "db": "Database",
                "host": "Host",
                "query_time": "Query time",
                "query_text": "Query text",
            }

            for rb_id, label in filter_options.items():
                if self.query_one(f"#filter_radio_buttons #{rb_id}", RadioButton).value:
                    filter_id = rb_id
                    filter_label = label
                    break

            if filter_label:
                if filter_id == "query_time":
                    if not modal_input.isdigit():
                        self.update_error_response("Query time must be an integer")
                    else:
                        self.dismiss([filter_label, int(modal_input)])
                elif filter_id == "query_text":
                    self.dismiss([filter_label, modal_input])
                else:
                    if filter_id == "host":
                        value = next((ip for ip, addr in self.host_cache_data.items() if modal_input == addr), None)
                        modal_input = value
                    else:
                        value = next(
                            (
                                getattr(data, filter_id)
                                for data in self.processlist_data.values()
                                if modal_input == getattr(data, filter_id)
                            ),
                            None,
                        )

                    self.dismiss([filter_label, modal_input])
            else:
                self.update_error_response("Please select a filter option")
        elif self.command == HotkeyCommands.thread_kill_by_parameter:
            kill_type = None
            lower_limit = None
            upper_limit = None

            checkbox_sleeping_queries = self.query_one("#sleeping_queries", Checkbox).value
            for rb in self.query("#kill_container RadioButton"):
                if rb.value:
                    kill_type = rb.id
                    break

            if kill_type:
                if kill_type == "time_range":
                    if re.search(r"(\d+-\d+)", modal_input):
                        time_range = modal_input.split("-")
                        lower_limit = int(time_range[0])
                        upper_limit = int(time_range[1])

                        if lower_limit > upper_limit:
                            self.update_error_response("Invalid time range! Lower limit can't be higher than upper")
                            return
                    else:
                        self.update_error_response("Invalid time range")
                        return

                self.dismiss([rb.id, modal_input, checkbox_sleeping_queries, lower_limit, upper_limit])
            else:
                self.update_error_response("Please select a kill option")
        elif self.command in [HotkeyCommands.thread_kill_by_id, HotkeyCommands.show_thread]:
            value = next((thread_id for thread_id in self.processlist_data.keys() if modal_input == thread_id), None)

            if not value:
                self.update_error_response(f"Thread ID [b red]{modal_input}[/b red] does not exist")
            else:
                self.dismiss(modal_input)
        elif self.command == HotkeyCommands.refresh_interval:
            if not modal_input.isnumeric():
                self.update_error_response("Input must be an integer")
                return

            modal_input = int(modal_input)
            if modal_input < 1:
                self.update_error_response("Input must be greater than 0")
                return

            self.dismiss(modal_input)
        else:
            self.dismiss(modal_input)

    def update_error_response(self, message):
        error_response = self.query_one("#error_response", Static)
        error_response.display = True
        error_response.update(message)



from dataclasses import dataclass
from typing import Dict, List

from dolphie.Modules.Functions import format_query, format_time
from dolphie.Modules.MySQL import Database
from rich.table import Table


@dataclass
class ConnectionStatus:
    connecting = "CONNECTING"
    disconnected = "DISCONNECTED"
    read_write = "R/W"
    read_only = "RO"


@dataclass
class Replica:
    thread_id: int
    host: str
    connection: Database = None
    table: Table = None
    replication_status: Dict[str, str] = None
    lag_source: str = None
    lag: int = None
    previous_sbm: int = 0
    mysql_version: str = None


class ReplicaManager:
    def __init__(self):
        self.available_replicas: list = []
        self.replicas: Dict[int, Replica] = {}
        self.ports: Dict[str, int] = {}

    def add(self, thread_id: int, host: str) -> Replica:
        self.replicas[thread_id] = Replica(thread_id=thread_id, host=host)

        return self.replicas[thread_id]

    def remove(self, thread_id: int):
        del self.replicas[thread_id]

    def get(self, thread_id: int) -> Replica:
        return self.replicas.get(thread_id)

    def remove_all(self):
        if self.replicas:
            for replica in self.replicas.values():
                if replica.connection:
                    replica.connection.close()

            self.replicas = {}

    def get_sorted_replicas(self) -> List[Replica]:
        return sorted(self.replicas.values(), key=lambda x: x.host)


@dataclass
class Panel:
    name: str
    visible: bool = False


class Panels:
    def __init__(self):
        self.dashboard = Panel("dashboard")
        self.processlist = Panel("processlist")
        self.graphs = Panel("graphs")
        self.replication = Panel("replication")
        # self.innodb_trx_locks = Panel("innodb_trx_locks")
        self.metadata_locks = Panel("metadata_locks")
        self.ddl = Panel("ddl")

    def get_panel(self, panel_name: str) -> Panel:
        return self.__dict__.get(panel_name, None)

    def get_all_panels(self) -> List[Panel]:
        return [panel for panel in self.__dict__.values() if isinstance(panel, Panel)]

    def all(self) -> List[str]:
        return [
            panel.name
            for name, panel in self.__dict__.items()
            if not name.startswith("__") and isinstance(panel, Panel)
        ]


class ProcesslistThread:
    def __init__(self, thread_data: Dict[str, str]):
        self.id = str(thread_data.get("id", ""))
        self.mysql_thread_id = thread_data.get("mysql_thread_id")
        self.user = thread_data.get("user", "")
        self.host = thread_data.get("host", "")
        self.db = thread_data.get("db", "")
        self.query = thread_data.get("query", "")
        self.time = int(thread_data.get("time", 0))
        self.protocol = self._get_formatted_string(thread_data.get("connection_type", ""))
        self.formatted_query = self._get_formatted_query(thread_data.get("query", ""))
        self.formatted_time = self._get_formatted_time()
        self.command = self._get_formatted_command(thread_data.get("command", ""))
        self.state = self._get_formatted_string(thread_data.get("state", ""))
        self.trx_state = self._get_formatted_string(thread_data.get("trx_state", ""))
        self.trx_operation_state = self._get_formatted_string(thread_data.get("trx_operation_state", ""))
        self.trx_rows_locked = self._get_formatted_number(thread_data.get("trx_rows_locked", 0))
        self.trx_rows_modified = self._get_formatted_number(thread_data.get("trx_rows_modified", 0))
        self.trx_concurrency_tickets = self._get_formatted_number(thread_data.get("trx_concurrency_tickets", 0))
        self.trx_time = self._get_formatted_trx_time(thread_data.get("trx_time", ""))

    def _get_formatted_time(self) -> str:
        thread_color = self._get_time_color()
        return f"[{thread_color}]{format_time(self.time)}[/{thread_color}]" if thread_color else format_time(self.time)

    def _get_time_color(self) -> str:
        thread_color = ""
        if "Group replication" not in self.query:  # Don't color GR threads
            if "SELECT /*!40001 SQL_NO_CACHE */ *" in self.query:
                thread_color = "purple"
            elif self.query:
                if self.time >= 10:
                    thread_color = "red"
                elif self.time >= 5:
                    thread_color = "yellow"
                else:
                    thread_color = "green"
        return thread_color

    def _get_formatted_command(self, command: str):
        return "[red]Killed[/red]" if command == "Killed" else command

    def _get_formatted_trx_time(self, trx_time: str):
        return format_time(int(trx_time)) if trx_time else "[dark_gray]N/A"

    def _get_formatted_query(self, query: str):
        return format_query(query)

    def _get_formatted_string(self, string: str):
        if not string:
            return "[dark_gray]N/A"

        return string

    def _get_formatted_number(self, number):
        if not number or number == "0":
            return "[dark_gray]0"

        return number


class HotkeyCommands:
    show_thread = "show_thread"
    thread_filter = "thread_filter"
    thread_kill_by_id = "thread_kill_by_id"
    thread_kill_by_parameter = "thread_kill_by_parameter"
    variable_search = "variable_search"
    rename_tab = "rename_tab"
    refresh_interval = "refresh_interval"



import argparse
import os
import re
import sys
from configparser import ConfigParser
from dataclasses import dataclass, field, fields
from typing import Dict, List
from urllib.parse import urlparse

import myloginpath
from dolphie.DataTypes import Panels
from dolphie.Modules.Queries import MySQLQueries
from rich import box
from rich.console import Console
from rich.table import Table
from rich.theme import Theme


@dataclass
class Config:
    app_version: str
    host_setup: bool = False
    user: str = None
    password: str = None
    host: str = "localhost"
    port: int = 3306
    socket: str = None
    ssl: Dict = field(default_factory=dict)
    ssl_mode: str = None
    ssl_ca: str = None
    ssl_cert: str = None
    ssl_key: str = None
    config_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/.dolphie.cnf")
    mycnf_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/.my.cnf")
    login_path: str = "client"
    host_cache_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/dolphie_host_cache")
    host_setup_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/dolphie_hosts")
    refresh_interval: int = 1
    heartbeat_table: str = None
    host_setup_available_hosts: List[str] = field(default_factory=list)
    startup_panels: str = "dashboard,processlist"
    graph_marker: str = "braille"
    pypi_repository: str = "https://pypi.org/pypi/dolphie/json"
    hostgroup: str = None
    hostgroup_hosts: Dict[str, List[str]] = field(default_factory=dict)
    show_trxs_only: bool = False
    show_additional_query_columns: bool = False
    # historical_trx_locks: bool = False


class ArgumentParser:
    def __init__(self, app_version: str):
        self.config_options = {}
        for variable in fields(Config):
            # Exclude these options since we handle them differently
            if variable.name not in [
                "app_version",
                "config_file",
                "host_setup_available_hosts",
                "ssl",
                "hostgroup_hosts",
            ]:
                self.config_options[variable.name] = variable.type

        self.formatted_options = "\n\t".join(
            [f"({data_type.__name__}) {option}" for option, data_type in self.config_options.items()]
        )
        epilog = f"""
Order of precedence for methods that pass options to Dolphie:
\t1. Command-line
\t2. Environment variables
\t3. Dolphie's config (set by --config-file)
\t4. ~/.mylogin.cnf (mysql_config_editor)
\t5. ~/.my.cnf (set by --mycnf-file)

MySQL my.cnf file supports these options under [client] section:
\thost
\tuser
\tpassword
\tport
\tsocket
\tssl_mode REQUIRED/VERIFY_CA/VERIFY_IDENTITY
\tssl_ca
\tssl_cert
\tssl_key

Login path file supports these options:
\thost
\tuser
\tpassword
\tport
\tsocket

Environment variables support these options:
\tDOLPHIE_USER
\tDOLPHIE_PASSWORD
\tDOLPHIE_HOST
\tDOLPHIE_PORT
\tDOLPHIE_SOCKET

Dolphie's config supports these options under [dolphie] section:
\t{self.formatted_options}
"""
        self.parser = argparse.ArgumentParser(
            conflict_handler="resolve",
            description="Dolphie, an intuitive feature-rich top tool for monitoring MySQL in real time",
            epilog=epilog,
            formatter_class=argparse.RawTextHelpFormatter,
        )
        self.config = Config(app_version)
        self.panels = Panels()

        self.console = Console(style="#e9e9e9", highlight=False)
        self.console.push_theme(
            Theme(
                {
                    "red2": "b #fb9a9a",
                }
            )
        )

        self._add_options()
        self._parse()

    def _add_options(self):
        self.parser.add_argument(
            "uri",
            metavar="uri",
            type=str,
            nargs="?",
            help=(
                "Use a URI string for credentials - format: mysql://user:password@host:port (port is optional with"
                f" default {self.config.port})"
            ),
        )

        self.parser.add_argument(
            "--host-setup",
            dest="host_setup",
            action="store_true",
            help="Start Dolphie by showing the Host Setup modal instead of automatically connecting",
        )
        self.parser.add_argument("-u", "--user", dest="user", type=str, help="Username for MySQL", metavar="")
        self.parser.add_argument("-p", "--password", dest="password", type=str, help="Password for MySQL", metavar="")
        self.parser.add_argument(
            "-h",
            "--host",
            dest="host",
            type=str,
            help="Hostname/IP address for MySQL",
            metavar="",
        )
        self.parser.add_argument(
            "-P",
            "--port",
            dest="port",
            type=int,
            help="Port for MySQL (socket has precendence)",
            metavar="",
        )
        self.parser.add_argument(
            "-S",
            "--socket",
            dest="socket",
            type=str,
            help="Socket file for MySQL",
            metavar="",
        )
        self.parser.add_argument(
            "--config-file",
            dest="config_file",
            type=str,
            help=(
                f"Dolphie's config file to use. Options are read from these files in the given order: "
                f"/etc/dolphie.cnf, {self.config.config_file}"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--mycnf-file",
            dest="mycnf_file",
            type=str,
            help=(
                "MySQL config file path to use. This should use [client] section "
                f"[default: {self.config.mycnf_file}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-f",
            "--host-cache-file",
            dest="host_cache_file",
            type=str,
            help=(
                "Resolve IPs to hostnames when your DNS is unable to. Each IP/hostname pair should be on its own line "
                f"using format ip=hostname [default: {self.config.host_cache_file}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-q",
            "--host-setup-file",
            dest="host_setup_file",
            type=str,
            help=(
                "Specify location of file that stores the available hosts to use in host setup modal [default:"
                f" {self.config.host_setup_file}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-l",
            "--login-path",
            dest="login_path",
            type=str,
            help=(
                "Specify login path to use with mysql_config_editor's file ~/.mylogin.cnf for encrypted login"
                f" credentials [default: {self.config.login_path}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-r",
            "--refresh_interval",
            dest="refresh_interval",
            type=int,
            help=f"How much time to wait in seconds between each refresh [default: {self.config.refresh_interval}]",
            metavar="",
        )
        self.parser.add_argument(
            "-H",
            "--heartbeat-table",
            dest="heartbeat_table",
            type=str,
            help=(
                "If your hosts use pt-heartbeat, specify table in format db.table to use the timestamp it "
                "has for replication lag instead of Seconds_Behind_Master from SHOW REPLICA STATUS"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-mode",
            dest="ssl_mode",
            type=str,
            help=(
                "Desired security state of the connection to the host. Supports: "
                "REQUIRED/VERIFY_CA/VERIFY_IDENTITY [default: OFF]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-ca",
            dest="ssl_ca",
            type=str,
            help="Path to the file that contains a CA (certificate authority)",
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-cert",
            dest="ssl_cert",
            type=str,
            help="Path to the file that contains a certificate",
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-key",
            dest="ssl_key",
            type=str,
            help="Path to the file that contains a private key for the certificate",
            metavar="",
        )
        self.parser.add_argument(
            "--panels",
            dest="startup_panels",
            type=str,
            help=(
                "What panels to display on startup separated by a comma. Supports:"
                f" {','.join(self.panels.all())} [default: {self.config.startup_panels}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--graph-marker",
            dest="graph_marker",
            type=str,
            help=(
                "What marker to use for graphs (available options: https://tinyurl.com/dolphie-markers) [default:"
                f" {self.config.graph_marker}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--pypi-repository",
            dest="pypi_repository",
            type=str,
            help=(
                "What PyPi repository to use when checking for a new version."
                " If not specified, it will use Dolphie's PyPi repository"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--hostgroup",
            dest="hostgroup",
            type=str,
            help=(
                "This is used for creating tabs and connecting to them for hosts you specify in"
                " Dolphie's config file under a hostgroup section. As an example, you'll have a section"
                " called [cluster1] then below it you will list each host on a new line in the format"
                " key=host (keys have no meaning). Hosts support optional port (default is whatever port parameter is)"
                " in the format host:port. You can also name the tabs by suffixing"
                " ~tab_name to the host (i.e. 1=host~tab_name)"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--show-trxs-only",
            dest="show_trxs_only",
            action="store_true",
            help="Start with only showing threads that have an active transaction",
        )
        self.parser.add_argument(
            "--additional-columns",
            dest="show_additional_query_columns",
            action="store_true",
            help="Start with additional columns in Processlist panel",
        )
        # self.parser.add_argument(
        #     "--historical-trx-locks",
        #     dest="historical_trx_locks",
        #     action="store_true",
        #     help=(
        #         "Always run the InnoDB TRX Locks query so it can save historical data to its graph instead of only "
        #         "when the panel is open. This query can be expensive in some environments"
        #     ),
        # )
        self.parser.add_argument(
            "--debug-options",
            dest="debug_options",
            action="store_true",
            help="Display options that are set and what they're set by (command-line, dolphie config, etc) then exit",
        )
        self.parser.add_argument(
            "-V",
            "--version",
            action="version",
            version=self.config.app_version,
            help="Display version and exit",
        )

    def set_config_value(self, source, option, value):
        setattr(self.config, option, value)

        if self.debug_options:
            self.debug_options_table.add_row(source, option, str(value))

    def _parse(self):
        login_options = ["user", "password", "host", "port", "socket"]

        options = vars(self.parser.parse_args())  # Convert object to dictionary

        dolphie_config_login_options_used = {}
        hostgroups = {}

        self.debug_options = False
        if options["debug_options"]:
            self.debug_options = True

            self.debug_options_table = Table(box=box.SIMPLE_HEAVY, header_style="b", style="#333f62")
            self.debug_options_table.add_column("Source")
            self.debug_options_table.add_column("Option", style="#91abec")
            self.debug_options_table.add_column("Value", style="#bbc8e8")

        config_files = ["/etc/dolphie.cnf", self.config.config_file]
        if options["config_file"]:
            config_files = [options["config_file"]]

        # Loop through config files to find the supplied options
        for config_file in config_files:
            if os.path.isfile(config_file):
                cfg = ConfigParser()
                cfg.read(config_file)

                # Loop through all of available options
                for option, data_type in self.config_options.items():
                    # If the option is in the config file
                    if cfg.has_option("dolphie", option):
                        # Check if the value is of the correct data type
                        value = self.verify_config_value(option, cfg.get("dolphie", option), data_type)

                        # Set the option to the value from the config file
                        self.set_config_value(f"dolphie config {config_file}", option, value)

                        # Save the login option to be used later
                        if option in login_options:
                            dolphie_config_login_options_used[option] = value

                # Save all hostgroups found to the config object
                for hostgroup in cfg.sections():
                    hosts = []
                    if hostgroup == "dolphie":
                        continue

                    for key in cfg.options(hostgroup):
                        host = cfg.get(hostgroup, key).strip()
                        if host:
                            hosts.append(host)
                        else:
                            self.exit(
                                f"{config_file}: Hostgroup [red2]{hostgroup}[/red2] has an empty host"
                                f" for key [red2]{key}[/red2]"
                            )

                    if not hosts:
                        self.exit(
                            f"{config_file}: Hostgroup [red2]{hostgroup}[/red2] cannot be loaded because"
                            f" it doesn't have any hosts listed under its section in Dolphie's config"
                        )

                    hostgroups[hostgroup] = hosts

        # Save the hostgroups found to the config object
        self.config.hostgroup_hosts = hostgroups

        # If the options are not set from Dolphie configs, set them to what command-line is or default
        for option in self.config_options.keys():
            value = getattr(self.config, option)

            # Override the option with command-line arguments if the option isn't for login
            if option not in login_options and options[option] and value != options[option]:
                self.set_config_value("command-line", option, options[option])

            # If the option is not set, set it to the default from Config's dataclass
            if not options[option]:
                options[option] = value

        # Use MySQL's my.cnf file for login options if specified
        if os.path.isfile(self.config.mycnf_file):
            cfg = ConfigParser()
            cfg.read(self.config.mycnf_file)

            for option in login_options:
                if cfg.has_option("client", option):
                    self.set_config_value("my.cnf", option, cfg.get("client", option))

            self.parse_ssl_options(cfg)

        # Use login path for login options if specified
        if self.config.login_path:
            try:
                login_path_data = myloginpath.parse(self.config.login_path)

                for option in login_options:
                    if option in login_path_data:
                        self.set_config_value("login path", option, login_path_data[option])
            except Exception as e:
                # Don't error out for default login path
                if self.config.login_path != "client":
                    self.exit(f"Problem reading login path file: {e}")

        # Update login options based on precedence
        for option in login_options:
            # Update config object with Dolphie config
            dolphie_value = dolphie_config_login_options_used.get(option)
            if dolphie_value is not None:
                self.set_config_value("dolphie config", option, dolphie_value)

            # Use environment variables if specified
            environment_var = f"DOLPHIE_{option.upper()}"
            env_value = os.environ.get(environment_var)
            if env_value is not None:
                self.set_config_value("environment", option, env_value)

            # Use command-line arguments if specified
            if options.get(option):
                self.set_config_value("command-line", option, options[option])

        # Lastly, parse URI if specified
        if options["uri"]:
            try:
                parsed = urlparse(options["uri"])

                if parsed.scheme != "mysql":
                    self.exit("Invalid URI scheme: Only 'mysql' is supported (see --help for more information)")

                self.config.user = parsed.username
                self.config.password = parsed.password
                self.config.host = parsed.hostname
                self.config.port = parsed.port or 3306
            except Exception as e:
                self.exit(f"Invalid URI: {e} (see --help for more information)")

        # Sanity check for hostgroup
        if self.config.hostgroup:
            if self.config.hostgroup not in hostgroups:
                self.exit(
                    f"Hostgroup [red2]{self.config.hostgroup}[/red2] cannot be used because"
                    f" it wasn't found in Dolphie's config"
                )

        if self.config.heartbeat_table:
            pattern_match = re.search(r"^(\w+\.\w+)$", self.config.heartbeat_table)
            if pattern_match:
                MySQLQueries.heartbeat_replica_lag = MySQLQueries.heartbeat_replica_lag.replace(
                    "$1", self.config.heartbeat_table
                )
            else:
                self.exit("Your heartbeat table did not conform to the proper format: db.table")

        self.parse_ssl_options(options)

        self.config.startup_panels = self.config.startup_panels.split(",")
        for panel in self.config.startup_panels:
            if panel not in self.panels.all():
                self.exit(f"Panel [red2]{panel}[/red2] is not valid (see --help for more information)")

        if os.path.exists(self.config.host_setup_file):
            with open(self.config.host_setup_file, "r") as file:
                self.config.host_setup_available_hosts = [line.strip() for line in file]

        if self.debug_options:
            self.console.print(self.debug_options_table)
            self.console.print("[#969aad]Note: Options are set by their source in the order they appear")
            sys.exit()

    def parse_ssl_options(self, data):
        if isinstance(data, ConfigParser):
            ssl_mode = data.get("client", "ssl_mode", fallback=None)
            ssl_ca = data.get("client", "ssl_ca", fallback=None)
            ssl_cert = data.get("client", "ssl_cert", fallback=None)
            ssl_key = data.get("client", "ssl_key", fallback=None)
        else:
            ssl_mode = data.get("ssl_mode")
            ssl_ca = data.get("ssl_ca")
            ssl_cert = data.get("ssl_cert")
            ssl_key = data.get("ssl_key")

        if ssl_mode:
            ssl_mode = ssl_mode.upper()

            if ssl_mode == "REQUIRED":
                self.config.ssl["required"] = True
            elif ssl_mode == "VERIFY_CA":
                if not ssl_ca:
                    self.exit("SSL mode [red2]VERIFY_CA[/red2] requires a CA file (--ssl-ca) to be specified")

                self.config.ssl["check_hostname"] = False
                self.config.ssl["verify_mode"] = True
            elif ssl_mode == "VERIFY_IDENTITY":
                if not ssl_ca:
                    self.exit("SSL mode [red2]VERIFY_IDENTITY[/red2] requires a CA file (--ssl-ca) to be specified")

                self.config.ssl["check_hostname"] = True
                self.config.ssl["verify_mode"] = True
            else:
                self.exit(f"Unsupported SSL mode [red2]{ssl_mode}[/red2]")

            if ssl_ca:
                self.config.ssl["ca"] = ssl_ca
            if ssl_cert:
                self.config.ssl["cert"] = ssl_cert
            if ssl_key:
                self.config.ssl["key"] = ssl_key

    def verify_config_value(self, option, value, data_type):
        if data_type == bool:
            if value.lower() == "true":
                return True
            elif value.lower() == "false":
                return False
            else:
                self.exit(
                    f"Error with Dolphie config: [red2]{option}[/red2] is a boolean and must either be true/false"
                )

        elif data_type == int:
            try:
                return int(value)
            except ValueError:
                self.exit(f"Error with Dolphie config: [red2]{option}[/red2] is an integer and must be a number")
        else:
            return value

    def exit(self, message):
        self.console.print(f"[indian_red]{message}[/indian_red]")
        sys.exit()
